import struct
import math

# This wrapper function keeps the file reading code tidy
def _struct_read(s, file):
    return s.unpack(file.read(s.size))

class ParseError(BaseException):
    pass
class Md3:
    # STRUCTS
    # A bunch of static objects to avoid repeated construction during loops
    header_struct = struct.Struct("<4sl64s9l")
    frame_struct = struct.Struct("<10f16s")
    tag_struct = struct.Struct("<64s12f")
    triangle_struct = struct.Struct("<3l")
    shader_struct = struct.Struct("<64sl")
    texcoord_struct = struct.Struct("<2f")
    vertex_struct = struct.Struct("<4h")
    surface_struct = struct.Struct("<4s64sh")

    # INIT
    # Fill in defaults everywhere to avoid AtrributeError and create any lists
    def __init__(self):
        # initialise all the exported properties to minimal defaults
        self.ident = b"IDP3"
        self.version = 15
        self.name = b""
        self.flags = 0
        self.frames = []
        self.tags = []
        self.surfaces = []
        

    # HEADER
    # This is stored against the body of the Md3 object
    # a) because you can only have a single header per model and
    # b) because quantities like num_frames only make sense in terms of an
    #   actual collection of skins belonging to the model
    # Note: private fields like _num_frames are only used to store the lengths
    # of field parsed from the header of a model file we are reading
    # Any code not reading from a binary file should get the size of collections
    # from len(self.frames) etc. as per the write_header function below
    def read_header(self, file):
        (ident, version,
         self.name, self.flags,
         self._num_frames,
         self._num_tags,
         self._num_surfaces,
         self._num_skins,
         self._ofs_frames,
         self._ofs_tags,
         self._ofs_surfaces,
         self._ofs_eof) = _struct_read(Md3.header_struct, file)
        #run some basic checks that we have been given a md3 file
        if ident != b"IDP3":
            raise ParseError("Ident does not match IDP3")
        if version != 15:
            raise ParseError("mdl file is not version 15")
        return self

    def write_header(self, file, ofs_frames=0, ofs_tags=0,
                     ofs_surfaces=0, ofs_eof=0):
        file.write(Md3.header_struct.pack
                   (
                       b"IDP3", 15,     
                       self.name, self.flags,
                       len(self.frames),
                       len(self.tags),
                       len(self.surfaces),
                       0,
                       ofs_frames,
                       ofs_tags,
                       ofs_surfaces,
                       ofs_eof
                    )
                   )
        return self

    class Frame:
        def __init__(self):
            self.origin = (0.0, 0.0, 0.0)
            self.name = b"nameless"
            
        def read(self, file):
            data = _struct_read(Md3.frame_struct, file)
            self._mins = data[0:3]
            self._maxs = data[3:6]
            self.origin = data[6:9]
            self._radius = data[9]
            self.name = data[10]
            return self

        def resize(self, surfaces):
            return self
        
        def write(self, file):
            file.write(Md3.frame_struct(self._mins[0],#FIXME calc actual bbox
                                        self._mins[1],
                                        self._mins[2],
                                        self._maxs[0],
                                        self._maxs[1],
                                        self._maxs[2],
                                        self.origin[0],
                                        self.origin[1],
                                        self.origin[2],
                                        self.radius,
                                        self.name)
                       )
            return self
        
        
    class Tag:
        def __init__(self):
            self.name = b""
            self.origin = (0.0, 0.0, 0.0)
            self.axis = ((1.0, 0.0, 0.0),(0.0, 1.0, 0.0),(0.0, 0.0, 1.0))

        def read(self, file):
            data = _struct_read(Md3.tag_struct, file)
            self.name = data[0]
            self.origin = data[1:4]
            self.axis=(data[4:7], data[7:10], data[10:13])
            return self

        def write(self, file):
            data = (self.name,) + self.origin
            data += self.axis[0] + self.axis[1] + self.axis[2]
            file.write(Md3.tag_struct.pack(*data))
            return self


    class Triangle:
        def __init__(self):
            self.vertices = (0, 0, 0)

        def read(self, file):
            self.vertices = _struct_read(Md3.triangle_struct, file)
            return self

        def write(self, file):
            file.write(Md3.triangle_struct.pack(self.vertices))
            return self

    class Shader:
        def __init__(self):
            self.name = b""
            self.index = 0

        def read(self, file):
            (self.name, self.index) = _struct_read(Md3.shader_struct, file)
            return self

        def write(self, file):
            file.write(Md3.shader_struct.pack(self.name, self.index))
            return self

    class Texcoord:
        def __init__(self):
            self.u, self.v = (0.0, 0.0)

        def read(self, file):
            self.u, self.v = _struct_read(Md3.texcoord_struct, file)
            return self

        def write(self, file):
            file.write(Md3.texcoord_struct.pack(self.u, self.v))
            return self

    class Vertex:
        def __init__(self):
            self.position = (0, 0, 0)
            self.normal = 0

        def read(self, file):
            data = _struct_read(Md3.vertex_struct, file)
            self.position = data[0:3]
            self.normal = data[3]
            return self

        def write(self, file):
            file.write(Md3.vertex_struct.pack(*self.position+self.normal))
            return self

        def encode(self, vector):
            x, y, z = vector
            lng = math.atan2(y, x) * 255 / (2 * math.pi)
            lat = math.acos(z) * 255 / (2 * math.pi)
            self.normal = (lng & 255) | (lat & 255) << 8

        def decode(self):
            lng = self.normal & 255
            lat = (self.normal >> 8) & 255
            x = math.cos(lat) * math.sin(lng)
            y = math.sin(lat) * math.sin(lng)
            z = math.cos(lng)
            return (x, y, z)

    class Surface:
        def __init__(self):
            self.name = b""
            self.flags =0
            self.shaders = []
            self.triangles = []
            self.texcoords = []
            self.vertices = []

        def read(self, file):
            surface_start = file.tell()
            (ident, self.name, self.flags,
             num_frames, num_shaders, num_vertices, num_triangles,
             ofs_triangles, ofs_shaders, ofs_texcoords, ofs_vertices, ofs_end
             ) = _struct_read(Md3.surface_struct, file)
            if ident != b"IDP3":
                raise ParseError("Surface ident does not match IDP3")
            # we must seek to the offset of each structure although if all
            # goes well these will be seeks of size 0 in most models
            file.seek(surface_start + ofs_shaders) #usually first
            for i in range(num_shaders):
                self.shaders.append(Md3.Shader().Read(file))
                
            file.seek(surface_start + ofs_triangles) 
            for i in range(num_triangles):
                self.triangles.append(Md3.Triangle().Read(file))
                
            file.seek(surface_start + ofs_texcoords) 
            for i in range(num_vertices):
                self.texcoords.append(Md3.Texcoord().Read(file))
                
            file.seek(surface_start + ofs_vertices) 
            for i in range(num_vertices * num_frames):
                self.vertices.append(Md3.Vertex().Read(file))
                
            # we must leave the file pointer at the end of structure given
            file.seek(surface_start + ofs_end)
            return self

        def write(self, file):
            num_frames = len(self.frames)
            num_shaders = len(self.shaders)
            num_vertices = len(self.vertices)
            num_triangles = len(self.triangles)
            ofs_shaders = Md3.shader_struct.size
            ofs_triangles = ofs_shaders+num_shaders*Md3.shader_struct.size
            ofs_texcoords = ofs_triangles+num_triangles*Md3.triangle_struct.size
            ofs_vertices = ofs_texcoords+num_vertices*Md3.texcoord_struct.size
            ofs_end = ofs_vertices+num_vertices*num_frames*Md3.vertex_struct.size
            file.write(Md3.shader_struct.pack(b"IDP3", self.name, self.flags,
             num_frames, num_shaders, num_vertices, num_triangles,
             ofs_triangles, ofs_shaders, ofs_texcoords, ofs_vertices, ofs_end))
            for shader in self.shaders:
                shader.write(file)
            for triangle in self.triangles:
                triangle.write(file)
            for texcoord in self.texcoords:
                texcoord.write(file)
            for vertex in self.vertices:
                vertex.write(file)

        
        
            
                
            
            
            
            

    # SKIN
    # Class representing a quake skin
    # Data is stored as a flat binary string
    # Object also has height and width of skin stored
    # 'pixel' helper function will return the palette index at x,y coordinates
    class Skin:
        def __init__(self, width, height):
            self.width = width
            self.height = height

        def read(self, file):
            self.pixels = file.read(self.width * self.height)
            return self

        def write(self, file, write_header=True):
            # Skingroups will call this on each skin they contain when written,
            # but with write_header set to false, since they write a single
            # header for the entire skingroup
            if write_header:
                file.write(Mdl.single_struct.pack(0))
            file.write(self.pixels)
            return self

        def pixel(self, x, y):
            x = x % self.width
            y = y % self.height
            return self.pixels[x + y * self.height]


    # SKINGROUP
    # For grouped animated skins
    # Comprises a header including a list of durations, followed by a list of
    # Skin objects.
    # Unenforced constraint: the list of durations and skins must be equal in
    # length or the mdl file will be invalid
    class SkinGroup:
        def __init__(self, width, height):
            self.width = width
            self.height = height
            self.duration = []
            self.skins = []

        def read(self, file):
            count = _struct_read(Mdl.single_struct, file)[0]
            for i in range(count):
                self.duration.append(
                    _struct_read(Mdl.duration_struct, file)[0])
            for i in range(count):
                self.skins.append(Mdl.Skin(self.width, self.height).read(file))
            return self

        def write(self, file):
            file.write(Mdl.single_struct.pack(1))
            file.write(Mdl.single_struct.pack(len(self.skins)))
            for dur in self.duration:
                file.write(Mdl.duration_struct.pack(dur))
            for s in self.skins:
                s.write(file, write_header=False)
            return self
        
    # READ/WRITE skins
    # These functions are methods of the main Mdl class
    # They read to/write from the skins list on the Mdl object
    # Note that the list can contain a mixture of Skin and Skingroup objects
    # When writing we can treat the members identically by calling each object's
    # own 'write' function
    def read_skins(self, file):
        for i in range(self._num_skins):
            type = _struct_read(Mdl.single_struct, file)[0]
            if type == 0:
                skin = Mdl.Skin(self.skinwidth, self.skinheight)
            else:
                skin = Mdl.SkinGroup(self.skinwidth, self.skinheight)
            self.skins.append(skin.read(file))
        return self

    def write_skins(self, file):
        for skin in self.skins:
            skin.write(file)
        return self
            

    # VERTEX
    # This class stores the skin coordinates for each vertex in the model
    # The vertex positions in the poses of the model are stored in the frame
    # classes. The triangles refer to the vertices by index of their position
    # in this list.
    # Unenforced constraints: The vertex list must remain in sync with the
    # triangle and frame list. For example: removing a vertex from this list
    # requires you to remove the vertex from every pose in the model frames.
    # In addition you must update the triangles list, not just to remove every
    # triangle which contained the vertex. You must also to reduce by one the
    # index values recorded in each triangle of every vertex which follows the
    # deleted vertex, since each will have shifted one place down the list.
    class Vertex:
        def read(self, file):
            self.onseam, self.u, self.v = _struct_read(Mdl.triple_struct, file)
            return self

        def write(self, file):
            file.write(Mdl.triple_struct.pack(self.onseam, self.u, self.v))
            return self
        
    # READ/WRITE VERTICES
    # These functions are methods of the main Mdl class
    # They operate on the 'vertices' list it maintains
    def read_vertices(self, file):
        for i in range(self._num_verts):
            self.vertices.append(Mdl.Vertex().read(file))
        return self

    def write_vertices(self, file):
        for vertex in self.vertices:
            vertex.write(file)
        return self

    # TRIANGLE
    # This class represents a triangle in the model as it is stored in the mdl
    # format, a backfacing flag followed by three indicies into the vertex list.
    # See VERTEX for the details of the relation between these classes and FRAME
    class Triangle:
        def read(self, file):
            self.backface = _struct_read(Mdl.single_struct, file)[0]
            self.vertices = _struct_read(Mdl.triple_struct, file)
            return self
            
        def write(self, file):
            file.write(Mdl.single_struct.pack(self.backface)+
                       Mdl.triple_struct.pack(*self.vertices)
                       )
            return self

    # READ/WRITE TRIANGLES
    # These functions are methods of the main Mdl class
    # They operate on the 'triangles' list it maintains
    def read_triangles(self, file):
        for i in range(self._num_tris):
            self.triangles.append(Mdl.Triangle().read(file))
        return self

    def write_triangles(self, file):
        for triangle in self.triangles:
            triangle.write(file)
        return self
            
    # COORD
    # This class stores the 'compressed' coordinate the mdl file uses for vertex
    # position. x, y, z values are stored as a byte in the range 0-255. Vertex
    # normals are stored in a 4th byte in the range 0-162. The normal value is
    # a lookup in the standard table of normals provided in anorms.h
    # This class contains a pair of helper functions to deal with the normals
    # 'decode' returns the normal vector that the coord stores
    # 'encode' takes a vector and sets the normal of the coord to the index of
    # the standard normal which best approximates its direction.
    # See also TRANSLATE below.
    class Coord:
        def read(self, file):
            coord = _struct_read(Mdl.coord_struct, file)
            self.position = coord[0:3]
            self.normal = coord[3]
            return self
            
        def write(self, file):
            file.write(Mdl.coord_struct.pack(*(self.position + (self.normal,))))
            return self
            
        def encode(self, vector):
            m = 0
            n = 0
            for i in range(len(normals)):
                #dot = sum(a * b for a, b in zip(vector, normals[i]))
                dot = vector[0] * normals[i][0] + vector[1] * normals[i][1] +\
                      vector[2] * normals[i][2] 
                if dot > m:
                    m = dot
                    n = i
            self.normal = n
            return self

        def decode(self):
            return normals[self.normal]
        
    # TRANSLATE
    # This is another helper function for a Coord, but it is a member of the Mdl
    # object. It translates the 0-255 range coords into floating point 'world
    # coordinates'. This method belongs to the Mdl itself because we need to use
    # the origin and scale properties of the Mdl to interpret the Coord.
    def translate(self, coord):
        return tuple(org + scale * x for org, scale, x in zip(
            self.origin, self.scale, coord.position))
            
    class Frame:
        def __init__(self):
            self.vertices = []
            
        def read(self, file, vertcount):
            self.bbox_min = Mdl.Coord()
            self.bbox_min.read(file)
            self.bbox_max = Mdl.Coord()
            self.bbox_max.read(file)
            self.name = file.read(16).split(b"\x00")[0]
            for i in range(vertcount):
                self.vertices.append(Mdl.Coord().read(file))
            return self
                
        def write(self, file, write_header=True):
            if write_header:
                file.write(Mdl.single_struct.pack(0))
            self.bbox_min.write(file)
            self.bbox_max.write(file)
            file.write(Mdl.name_struct.pack(self.name))
            for vert in self.vertices:
                vert.write(file)
            return self
                
    class FrameGroup:
        def __init__(self):
            self.duration = []
            self.frames = []
                          
        def read(self, file, vertcount):
            count = _struct_read(Mdl.single_struct, file)[0]
            self.sequencemin = Mdl.Coord()
            self.sequencemin.read(file)
            self.sequencemax = Mdl.Coord()
            self.sequencemax.read(file)
            for i in range(count):
                self.duration.append(_struct_read(Mdl.duration_struct, file)[0])
            for i in range(count):
                self.frames.append(Mdl.Frame().read(file, vertcount))
            return self

        def write(self, file):
            file.write(Mdl.single_struct.pack(1))
            file.write(Mdl.single_struct.pack(len(self.frames)))
            self.sequencemin.write(file)
            self.sequencemax.write(file)
            for dur in self.duration:
                file.write(Mdl.duration_struct.pack(dur))
            for frame in self.frames:
                frame.write(file, write_header=False)
            return self


    def read_frames(self, file):
        for i in range(self._num_frames):
            type = _struct_read(Mdl.single_struct, file)[0]
            if type == 0:
                frame = Mdl.Frame()
            else:
                frame = Mdl.FrameGroup()
            self.frames.append(frame.read(file, self._num_verts))
        return self
            
    def write_frames(self, file):
        for frame in self.frames:
            frame.write(file)
        return self


    def read(self, file):
        self.read_header(file)
        self.read_skins(file)
        self.read_vertices(file)
        self.read_triangles(file)
        self.read_frames(file)
        return self
        
    def write(self, file):
        self.write_header(file)
        self.write_skins(file)
        self.write_vertices(file)
        self.write_triangles(file)
        self.write_frames(file)
        return self


